# This Python file uses the following encoding: utf-8
# Copyright (c) 2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' pytest module for the Fortran2003 Do Construct - R826.'''

import pytest
from fparser.api import get_reader
from fparser.common.readfortran import FortranStringReader
from fparser.two.Fortran2003 import Block_Label_Do_Construct, \
    Block_Nonlabel_Do_Construct


@pytest.mark.usefixtures("f2003_create")
def test_block_label_do_construct():
    ''' Tests that block labeled DO construct
    is parsed correctly (R826_1). '''
    tcls = Block_Label_Do_Construct

    obj = tcls(get_reader('''\
      do 12
        a = 1
 12   continue
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'DO 12\n  a = 1\n12 CONTINUE'

    obj = tcls(get_reader('''\
      foo: do 21, i=1,10
        a = 1
 21   end do foo
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'foo:DO 21 , i = 1, 10\n  a = 1\n21 END DO foo'

    obj = tcls(get_reader('''
      do 51 while (a < 10)
        a = a + 1
 51   continue
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'DO 51 WHILE (a < 10)\n  a = a + 1\n51 CONTINUE'

    obj = tcls(get_reader('''
      do 52
        a = a + 1
        if (a > 10) exit
 52   continue
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'DO 52\n  a = a + 1\n  IF (a > 10) EXIT\n52 CONTINUE'

    obj = tcls(get_reader('''\
      do 12
        do 13
          a = 1
 13   continue
 12   continue
    '''))
    assert str(obj) == 'DO 12\n  DO 13\n    a = 1\n13 CONTINUE\n12 CONTINUE'
    assert len(obj.content) == 3, repr(len(obj.content))
    assert str(obj.content[1]) == 'DO 13\n  a = 1\n13 CONTINUE'

    obj = tcls(get_reader('''
      do 52, i = 1,10
        do 53, while (j /= n)
        j = j + i
 53   continue
 52   continue
    '''))
    assert len(obj.content) == 3, repr(len(obj.content))
    assert str(obj) == ('DO 52 , i = 1, 10\n  DO 53 , WHILE (j /= n)\n'
                        '    j = j + i\n53 CONTINUE\n52 CONTINUE')
    assert str(obj.content[1]) == (
        'DO 53 , WHILE (j /= n)\n  j = j + i\n53 CONTINUE')


@pytest.mark.usefixtures("f2003_create")
def test_block_nonlabel_do_construct():
    # pylint: disable=invalid-name
    ''' Tests that block nonlabeled DO construct is parsed
    correctly (R826_2) '''
    tcls = Block_Nonlabel_Do_Construct

    obj = tcls(get_reader('''\
      do i=1,10
        a = 1
      end do
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'DO i = 1, 10\n  a = 1\nEND DO'

    obj = tcls(get_reader('''\
      do while (a < 10)
        a = a + 1
      end do
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'DO WHILE (a < 10)\n  a = a + 1\nEND DO'

    obj = tcls(get_reader('''
      do
        a = a - 1
        if (a < 10) exit
      end do
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'DO\n  a = a - 1\n  IF (a < 10) EXIT\nEND DO'
    assert len(obj.content) == 4, repr(len(obj.content))
    assert str(obj.content[2]) == 'IF (a < 10) EXIT'

    obj = tcls(get_reader('''\
      foo:do i=1,10
        a = 1
      end do foo
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'foo:DO i = 1, 10\n  a = 1\nEND DO foo'

    obj = tcls(get_reader('''\
      foo:do while (a < 10)
        a = a + 1
      end do foo
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'foo:DO WHILE (a < 10)\n  a = a + 1\nEND DO foo'

    obj = tcls(get_reader('''\
      do j=1,2
      foo:do i=1,10
        a = 1
      end do foo
      end do
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == ('DO j = 1, 2\n'
                        '  foo:DO i = 1, 10\n    a = 1\n  END DO foo\nEND DO')

    obj = tcls(get_reader('''
      do while (j >= n)
      bar:do i=1,10
        a = i + j
      end do bar
      j = j - 1
      end do
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == ('DO WHILE (j >= n)\n'
                        '  bar:DO i = 1, 10\n    a = i + j\n  END DO bar\n'
                        '  j = j - 1\nEND DO')

    obj = tcls(get_reader('''
      do, i = 1,10
      bar: do, while (j /= n)
        a = i - j
      end do bar
      end do
    '''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == (
        'DO , i = 1, 10\n'
        '  bar:DO , WHILE (j /= n)\n    a = i - j\n  END DO bar\n'
        'END DO')
    assert len(obj.content) == 3, repr(len(obj.content))
    assert str(obj.content[1]) == (
        'bar:DO , WHILE (j /= n)\n  a = i - j\nEND DO bar')


@pytest.mark.usefixtures("f2003_create")
def test_doconstruct_tofortran_non_ascii():
    ''' Check that the tofortran() method works when the do-construct contains
    a character string with non-ascii characters. '''
    code = (u"DO i=1,10\n"
            u"  WRITE(*,*) ' for e1=1\xb0'\n"
            u"END DO\n")
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Block_Nonlabel_Do_Construct(reader)
    out_str = str(obj)
    assert "for e1=1" in out_str



# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# --------------------------------------------------------------------
#
# The original software (in the f2py project) was distributed under
# the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Test parsing of whole fortran files; 'blackbox' tests here.
"""

from os.path import join, dirname
from fparser import api

SOURCE_STR = '''\
    ! before foo
    subroutine foo
    integer i, r
    do i = 1,100
      r = r + 1
    end do
    ! after end do
    end subroutine foo
    '''


def test_use_module():
    '''
    Test that we parse a function that uses a module when we supply the
    latter via the `source_only` argument.
    '''
    cwd = dirname(__file__)
    sources = [join(cwd, 'modfile.f95'), join(cwd, 'funcfile.f95')]
    file_to_parse = sources[1]
    _ = api.parse(file_to_parse, isfree=True, isstrict=False,
                  source_only=sources)


def test_dimension_attr():
    '''
    Tests for parsing the `dimension` attribute of variable declarations.
    '''
    source_str = ("    subroutine foo\n"
                  "    integer, dimension( -10 : 10, -   2147483648 : "
                  "2147483648) :: a( -2 : 2, 1000000 : 1000001 )\n"
                  "    real, dimension(-20:20, 100:113, -  512  : 713) :: b\n"
                  "    end\n")
    tree = api.parse(source_str, isfree=True, isstrict=False)
    subr = tree.a.external_subprogram['foo']
    avar = subr.a.variables['a']

    assert avar.dimension == [('-10', '10'), ('-   2147483648', '2147483648')]
    assert avar.bounds == [('-2', '2'), ('1000000', '1000001')]
    assert avar.shape == ['5', '2']

    bvar = subr.a.variables['b']

    assert bvar.dimension == [('-20', '20'), ('100', '113'), ('-  512', '713')]
    assert bvar.shape == ['41', '14', '1226']


def test_provides():
    '''
    Tests for non-Fortran (f2py markup) module_provides and use_provides
    key words. This is a legacy feature that is due to fparser's origins
    in the f2py project (https://docs.scipy.org/doc/numpy/f2py/).
    '''
    source_str = '''
    module mod1
    implicit none
    integer, parameter :: GP = 6
    integer :: a,b,c,d,e
    ! module_provides = {GP,a,b,c,d,e}
    ! use_provides = {}
    end module mod1

    module mod2
    implicit none
    integer, parameter :: SP = 5
    real :: a,b,c
    ! module_provides = {SP,a,b,c}
    ! use_provides = {}
    end module mod2

    module mod3
    use mod1
    implicit none
    integer, parameter :: DP = 0
    ! module_provides = {DP}
    ! use_provides = {GP,a,b,c,d,e}
    end module mod3

    module mod4
    use mod2
    implicit none
    ! module_provides = {}
    ! use_provides = {SP,a,b,c}
    end module mod4

    module mod5
    use mod3, only: lGP => GP, a,b,e
    use mod4, only: a2 => a, b2 => b
    implicit none

    integer, parameter :: FP = 1000
    integer(kind=kind(0)) :: dummy
    parameter (dummy = 20)
    integer, private :: x,y,z

    ! module_provides = {FP, dummy}
    ! use_provides = {lGP, a, b, e, a2, b2}
    end module mod5

    module mod6
    use mod5, qgp => lgp
    implicit none
    ! module_provides = {}
    ! use_provides = {FP, dummy, a2, b2, qgp, a, b, e}
    end module mod6

      '''

    tree = api.parse(source_str, isfree=True, isstrict=False)
    mod5 = tree.a.module['mod5']
    mod6 = tree.a.module['mod6']
    assert list(mod5.a.module_provides.keys()).sort() == ['fp', 'dummy'].sort()
    assert list(mod5.a.use_provides.keys()).sort() == \
        ['a', 'b', 'e', 'a2', 'b2', 'lgp'].sort()
    assert list(mod6.a.module_provides.keys()) == []
    assert list(mod6.a.use_provides.keys()).sort() == \
        ['fp', 'dummy', 'b', 'e', 'qgp', 'a2', 'a', 'b2'].sort()
    assert mod6.a.use_provides['qgp'].name == 'gp'


def test_walk():
    '''
    Test the walk() method of the api module.
    '''
    tree = api.parse(SOURCE_STR, isfree=True, isstrict=False,
                     ignore_comments=False)
    for stmt, depth in api.walk(tree, 1):
        print(depth, stmt.item)


def test_caching():
    ''' Tests relating to enabling/disabling of parser cache. '''
    tree1 = api.parse(SOURCE_STR, isfree=True, isstrict=False,
                      ignore_comments=False)
    # If we keep the cache from the last call to parse then we
    # should get the same object back
    tree2 = api.parse(SOURCE_STR, isfree=True, isstrict=False,
                      ignore_comments=False, clear_cache=False)
    assert tree1 is tree2
    # Now wipe the cache and check that we get a new object
    tree2 = api.parse(SOURCE_STR, isfree=True, isstrict=False,
                      ignore_comments=False, clear_cache=True)
    assert tree1 is not tree2
    # Check that wiping the cache is the default behaviour
    tree3 = api.parse(SOURCE_STR, isfree=True, isstrict=False,
                      ignore_comments=False)
    assert tree3 is not tree2



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R1101 : This file tests the support for a
main program with a program statement. It does not test the case where
there is no program statement. That situation is covered by the
Main_Program0 class. Tests for Main_Program0 are currently in
test_fortran2003.py.

As this rule is about ordering of other rules we only need to test the
different combinations. We don't need to test the content of
individual rules (e.g we do not need to worry about "end", "end
program", "end program a" as these should be covered by the tests for
End_Program_Unit

'''

import pytest
from fparser.two.utils import FortranSyntaxError, NoMatchError
from fparser.api import get_reader
from fparser.two.Fortran2003 import Main_Program


def test_valid(f2003_create):
    ''' Test that valid code is parsed correctly. '''

    # basic
    obj = Main_Program(get_reader("program a\nend"))
    assert isinstance(obj, Main_Program)
    assert str(obj) == 'PROGRAM a\nEND'
    assert repr(obj) == ("Main_Program(Program_Stmt('PROGRAM', Name('a')), "
                         "End_Program_Stmt(None, None))")

    # name matching
    obj = Main_Program(get_reader("program a\nend program a"))
    assert isinstance(obj, Main_Program)
    assert str(obj) == 'PROGRAM a\nEND PROGRAM a'
    assert repr(obj) == ("Main_Program(Program_Stmt('PROGRAM', Name('a')), "
                         "End_Program_Stmt('PROGRAM', Name('a')))")

    # mixed case name matching
    obj = Main_Program(get_reader("program a\nend program A"))
    assert isinstance(obj, Main_Program)
    assert str(obj) == 'PROGRAM a\nEND PROGRAM A'

    # specification-part
    obj = Main_Program(get_reader("program a\ninteger i\nend program a"))
    assert str(obj) == 'PROGRAM a\n  INTEGER :: i\nEND PROGRAM a'

    # execution-part
    obj = Main_Program(get_reader("program a\ni=10\nend program a"))
    assert str(obj) == 'PROGRAM a\n  i = 10\nEND PROGRAM a'

    # internal-subprogram-part
    obj = Main_Program(get_reader("program a\ncontains\nsubroutine foo\n"
                                  "end\nend program a"))
    assert str(obj) == ("PROGRAM a\n  CONTAINS\n  SUBROUTINE foo\n"
                        "  END\nEND PROGRAM a")

    # specification-part + execution-part
    obj = Main_Program(get_reader("program a\ninteger i\ni=10\nend program a"))
    assert str(obj) == 'PROGRAM a\n  INTEGER :: i\n  i = 10\nEND PROGRAM a'

    # execution-part + internal-subprogram-part
    obj = Main_Program(get_reader("program a\ni=10\ncontains\nsubroutine foo\n"
                                  "end\nend program a"))
    assert str(obj) == ("PROGRAM a\n  i = 10\n  CONTAINS\n  SUBROUTINE foo\n"
                        "  END\nEND PROGRAM a")

    # specification-part + execution-part + internal-subprogram-part
    obj = Main_Program(get_reader("program a\ninteger i\ni=10\ncontains\n"
                                  "subroutine foo\nend\nend program a"))
    assert str(obj) == ("PROGRAM a\n  INTEGER :: i\n  i = 10\n  CONTAINS\n  "
                        "SUBROUTINE foo\n  END\nEND PROGRAM a")


def test_invalid1(f2003_create):
    ''' Test that exceptions are raised for invalid code '''

    # no end
    with pytest.raises(NoMatchError) as excinfo:
        _ = Main_Program(get_reader("program a\n"))
    assert "at line 1\n>>>program a" in str(excinfo.value)

    # no start
    with pytest.raises(NoMatchError) as excinfo:
        _ = Main_Program(get_reader("end program a\n"))
    assert "at line 1\n>>>end program a" in str(excinfo.value)

    # name mismatch
    with pytest.raises(FortranSyntaxError) as excinfo:
        _ = Main_Program(get_reader("program a\nend program b"))
    assert "at line 2\n>>>end program b\nExpecting name 'a'" \
        in str(excinfo.value)


def test_invalid2(f2003_create):
    '''Test that specification-part after execution-part produces an
    error.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Main_Program(get_reader("program a\ni=10\ninteger i\n"
                                    "end program a"))
    assert "at line 3\n>>>integer i\n" in str(excinfo.value)


def test_invalid3(f2003_create):
    '''Test that execution-part after internal-subprogram-part produces an
    error.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Main_Program(get_reader("program a\ncontains\nsubroutine foo\n"
                                    "end\ni=10\nend program a"))
    assert "at line 5\n>>>i=10\n" in str(excinfo.value)



# Copyright (c) 2019-2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R438: this module contains pytest tests for the
   support for the components of a derived type. Note that only condition
   C449 is tested for here. The other conditions (C436-C452) are untested,
   this is the subject of #232.
'''

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Component_Part
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("var_type", ["integer",
                                      "logical",
                                      "character(len=1)",
                                      "real*8",
                                      "real(r_def)"])
def test_data_component_part(var_type):
    ''' Test that various data component declarations are
    recognised. R440. '''
    code = var_type + " :: iflag"
    reader = get_reader(code, isfree=True, isstrict=True)
    obj = Component_Part(reader)
    assert "iflag" in str(obj)


@pytest.mark.usefixtures("f2003_create")
def test_invalid_data_component_part():
    ''' Check that we don't get a match for an invalid component decln. '''
    code = "log :: iflag"
    reader = get_reader(code)
    with pytest.raises(NoMatchError):
        _ = Component_Part(reader)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("interface, attributes",
                         [("", "pointer, nopass"),
                          ("real_func", "pointer"),
                          ("real_func", "pointer, pass")])
def test_proc_component_part(interface, attributes):
    ''' Test that various procedure component declarations are
    recognised. R445. '''
    code = "procedure({0}), {1} :: my_proc".format(interface, attributes)
    reader = get_reader(code)
    obj = Component_Part(reader)
    assert obj is not None


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("invalid_code", ["procure(), nopass :: my_proc",
                                          "procedure, nopass :: my_proc",
                                          "procedure), nopass :: my_proc",
                                          "procedure(, nopass :: my_proc",
                                          "procedure() nopass :: my_proc",
                                          "procedure(), nopass my_proc"])
def test_invalid_proc_component(invalid_code):
    ''' Check that we don't get a match for an invalid procedure
    declaration. '''
    reader = get_reader(invalid_code)
    with pytest.raises(NoMatchError):
        Component_Part(reader)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("attributes", ["nopass", "pass"])
def test_proc_component_pointer(attributes):
    ''' R445, C449: POINTER shall appear in each
    proc-component-attr-spec-list. '''
    code = "procedure(), {0} :: my_proc".format(attributes)
    reader = get_reader(code)
    with pytest.raises(NoMatchError):
        _ = Component_Part(reader)



# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from fparser import api

def test_reproduce_issue():
    source_str = '''\
      subroutine gwinput_v2x(ifi,konf,ncore)
      integer :: ifi, !File handle. Write
     &  konf(0:lmxamx,nclass),! Principle
                              ! For examp
                              ! Core orbi
                              !   1, 2,..
                              !   2, 3,..
                              !   3, 4,..
                              !
     &  ncore(nclass)   ! ncore = \sum_l 
                        ! Number of diffe
      end
'''
    tree = api.parse(source_str, isfree=False, isstrict=False,
                     ignore_comments=False)
    assert str(tree).strip().split('\n')[1:]=='''
      !      BEGINSOURCE <cStringIO.StringI object at 0x1e52ea0> mode=fix90
        SUBROUTINE gwinput_v2x(ifi, konf, ncore)
          INTEGER ifi, konf(0:lmxamx,nclass), ncore(nclass)
          !File handle. Write
          ! Principle
          ! For examp
          ! Core orbi
          !   1, 2,..
          !   2, 3,..
          !   3, 4,..
          !
          ! ncore = \sum_l
          ! Number of diffe
        END SUBROUTINE gwinput_v2x
    '''.strip().split('\n')[1:]

def test_Issue_r25_1():
    source_str = '''
      subroutine bndfp()
      logical:: mlog
c test abb
c
#1232       
!

c
cabi
      real(8):: abc,a(5),
     &  abcx  !hhhh1
      abc=3   !hhhh2
      do i=1,5
         a(i)=i
      enddo
!

!
      end
    '''
    tree = api.parse(source_str, isfree=False, isstrict=False,
                     ignore_comments=False)
    assert str(tree).strip().split('\n')[1:]=='''
      !      BEGINSOURCE <cStringIO.StringI object at 0xb52b40> mode=fix90

        SUBROUTINE bndfp()
          LOGICAL mlog
          ! test abb
          !
          !1232
          !

          !
          !abi
          REAL(KIND=8) abc, a(5), abcx
          !hhhh1
          abc = 3
          !hhhh2
          DO  i=1,5
            a(i) = i
          END DO 
          !

          !
        END SUBROUTINE bndfp
    '''.strip().split('\n')[1:]

def test_Issue_r25_1():
    source_str = """
      subroutine rsedit()
        print 311
  311   format(/
     .  t4,'read [fn]',t15,

     .  t14,'  Third argument = flip: exchange n+, n- ')
      end

    """
    tree = api.parse(source_str, isfree=False, isstrict=False,
                     ignore_comments=False)
    assert str(tree).strip().split('\n')[1:]=="""
      !      BEGINSOURCE <cStringIO.StringI object at 0x121bab0> mode=fix90

        SUBROUTINE rsedit()
          PRINT 311
 311      FORMAT (/  t4, 'read [fn]', t15, t14, '  Third argument = flip: exchange n+, n- ')

        END SUBROUTINE rsedit
    """.strip().split('\n')[1:]

def test_comment_4():
    source_str = """
      subroutine m_struc_def()
 !abc test
      end
    """
    tree = api.parse(source_str, isfree=False, isstrict=False,
                     ignore_comments=False)
    assert str(tree).strip().split('\n')[1:]=="""
      !      BEGINSOURCE <cStringIO.StringI object at 0x121aa80> mode=fix90

        SUBROUTINE m_struc_def()
          !abc test
        END SUBROUTINE m_struc_def
""".strip().split('\n')[1:]




# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 Cray-pointers: This file tests the support for the
Cray-pointer statement.

'''

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Cray_Pointer_Stmt
from fparser.two.utils import NoMatchError


def test_cray_pointer_stmt(f2003_create):
    '''Check that a basic Cray-pointer statement is parsed
    correctly. Input separately as a string and as a reader object

    '''
    def check_use(reader):
        '''Internal helper function to avoid code replication.'''
        ast = Cray_Pointer_Stmt(reader)
        assert "POINTER(a, b)" in str(ast)
        assert (repr(ast).replace('u', '') ==
                "Cray_Pointer_Stmt('POINTER', Cray_Pointer_Decl_List(',', "
                "(Cray_Pointer_Decl(Name('a'), Name('b')),)))")

    line = "pointer (a, b)"
    check_use(line)
    reader = get_reader(line)
    check_use(reader)


def test_spaces(f2003_create):
    '''Check that spaces are allowed.'''
    line = "  pointer  ( a , b )  "
    ast = Cray_Pointer_Stmt(line)
    assert "POINTER(a, b)" in str(ast)


def test_case(f2003_create):
    '''Check that different case is allowed.'''
    line = "PoInTeR (a, b)"
    ast = Cray_Pointer_Stmt(line)
    assert "POINTER(a, b)" in str(ast)


def test_list(f2003_create):
    '''Check that a list of Cray-pointers is supported.'''
    line = "pointer (a, b), (c, d(1:n)), (e, f)"
    ast = Cray_Pointer_Stmt(line)
    assert "POINTER(a, b), (c, d(1 : n)), (e, f)" in str(ast)


def test_errors(f2003_create):
    '''Check that syntax errors produce a NoMatchError exception.'''
    for line in ["", "  ", "ponter (a, b)", "pointer", "pointer a, b"
                 "pointer (a, b) (a, b)"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Cray_Pointer_Stmt(line)
        assert "Cray_Pointer_Stmt: '{0}'".format(line) in str(excinfo.value)


def test_invalid_cray_pointer(f2003_create, monkeypatch):
    '''Test that the cray-pointer extension to the standard raises an
    exception if it is not named as a valid extension.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", [])
    myinput = "pointer (mypointer, mypointee)"
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cray_Pointer_Stmt(myinput)
        assert "Cray_Pointer_Stmt: '{0}'".format(myinput) \
            in str(excinfo.value)


def test_valid_cray_pointer(f2003_create, monkeypatch):
    '''Test that the cray-pointer extension to the standard produces the
    expected output if it is named as a valid extension.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["cray-pointer"])
    myinput = "pointer(mypointer, mypointee)"
    result = Cray_Pointer_Stmt(myinput)
    assert str(result).lower() == myinput



# Copyright (c) 2018-2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Module which provides pytest fixtures for use by files in this
directory

'''
import pytest
from fparser.two.parser import ParserFactory


@pytest.fixture
def f2003_create():
    '''Create a fortran 2003 parser class hierarchy'''
    _ = ParserFactory().create(std="f2003")


@pytest.fixture
def f2003_parser():
    '''Create a Fortran 2003 parser class hierarchy and return the parser
    for usage in tests.

    :return: a Program class (not object) for use with the Fortran reader.
    :rtype: :py:class:`fparser.two.Fortran2003.Program`
    '''
    return ParserFactory().create(std='f2003')



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R1002 : This file tests the support for a
format specification.

'''

import pytest
from fparser.two.Fortran2003 import Format_Specification
from fparser.two.utils import NoMatchError


def test_single(f2003_create):
    '''Check that some basic format specifications with single format
    items are parsed correctly. Individual items are tested by the
    associated classes.

    '''
    # 1: data-edit-desc, 2: r data-edit-desc, 3: control-edit-desc, 4:
    # char-string-edit-desc, 5: format-item-list
    for my_input in ["()", "(E2.2)", "(2E2.2)", "(/)", "('hello')",
                     "(2(E2.2))"]:
        ast = Format_Specification(my_input)
        assert str(ast) == my_input


def test_multi(f2003_create):
    '''Check that a basic format specification with multiple format items
    are parsed correctly. Individual format items are tested by the
    associated classes.

    '''
    # 1: data-edit-desc, 2: r data-edit-desc, 3: control-edit-desc, 4:
    # char-string-edit-desc, 5: format-item-list
    my_input = "(E2.2, 2E2.2, /, 'hello', 2(E2.2, 'there'))"
    ast = Format_Specification(my_input)
    assert str(ast) == my_input


def test_spaces(f2003_create):
    '''Check that a basic format specification with multiple format items
    are parsed correctly with spaces. Individual format items are
    tested by the associated classes. Note, the standard states that
    spaces can occur anywhere within a format specifier and not affect
    the meaning, other than within a format string. This is checked by
    the individual format item classes.

    '''
    # spaces before and after the brackets
    my_input = ("  (  )  ")
    ast = Format_Specification(my_input)
    assert str(ast) == "()"
    # 1: data-edit-desc, 2: r data-edit-desc, 3: control-edit-desc, 4:
    # char-string-edit-desc, 5: format-item-list
    my_input = ("(  E2.2  ,  2E2.2  ,  /  ,  'hello'  ,   "
                "2(E2.2, 'there')  )")
    ast = Format_Specification(my_input)
    assert str(ast) == "(E2.2, 2E2.2, /, 'hello', 2(E2.2, 'there'))"


def test_hollerith(f2003_create, monkeypatch):
    '''Check that a basic format specification with a hollerith string
    item is parsed correctly. Individual format items are tested by
    the associated classes.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    my_input = ("(2H,,)")
    ast = Format_Specification(my_input)
    assert str(ast) == my_input


def test_c1002(f2003_create, monkeypatch):
    '''Check that format specifications conforming to the C1002
    constraints are parsed correctly. This test actually checks class
    Format_Item_C1002 when a comma is missing as the constraints have
    been implemented within this class in fparser.

    '''
    # Comma is optional between a P edit descriptor and an immediately
    # following F, E, EN, ES, D, or G edit descriptor, possibly
    # preceded by a repeat specifier.
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    for specifier in ['F', 'E', 'EN', 'ES', 'D', 'G']:
        # Without repeat specifier.
        for my_input in ["(2P, {0}2.2)".format(specifier),
                         "(2P {0}2.2)".format(specifier),
                         "(2P{0}2.2)".format(specifier)]:
            ast = Format_Specification(my_input)
            assert str(ast) == "(2P, {0}2.2)".format(specifier)
        # With repeat specifier.
        for my_input in ["(2P, 2{0}2.2)".format(specifier),
                         "(2P 2{0}2.2)".format(specifier),
                         "(2P2{0}2.2)".format(specifier)]:
            ast = Format_Specification(my_input)
            assert str(ast) == "(2P, 2{0}2.2)".format(specifier)

    # spaces
    for my_input in [" ( 2 2 P , / ) ", " ( 2 2 P / ) ", " ( 2 2 P / ) "]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(22P, /)"

    # Comma is optional before a slash edit descriptor when the
    # optional repeat specification is not present.
    # data-edit-desc
    for my_input in ["(2P, /)", "(2P /)", "(2P/)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(2P, /)"
    # control-edit-desc
    for my_input in ["(2E2.2, /)", "(2E2.2 /)", "(2E2.2/)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(2E2.2, /)"
    # self
    for my_input in ["(/, /)", "(/ /)", "(//)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(/, /)"
    # char-string-edit-desc
    for my_input in ["('hello', /)", "('hello' /)", "('hello'/)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "('hello', /)"
    # format-item-list
    for my_input in ["(2(/'hello'), /)", "(2(/'hello') /)", "(2(/'hello')/)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(2(/, 'hello'), /)"
    # Hollerith string
    for my_input in ["(3H2.2, /)", "(3H2.2 /)", "(3H2.2/)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(3H2.2, /)"

    # Comma is optional after a slash edit descriptor.
    # data-edit-desc
    for my_input in ["(/, 2P)", "(/ 2P)", "(/2P)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(/, 2P)"
    # data-edit-desc (/ with repeat specification)
    for my_input in ["(2/, 2P)", "(2/ 2P)", "(2/2P)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(2/, 2P)"
    # control-edit-desc
    for my_input in ["(/, 2E2.2)", "(/ 2E2.2)", "(/2E2.2)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(/, 2E2.2)"
    # char-string-edit-desc
    for my_input in ["(/, 'hello')", "(/ 'hello')", "(/'hello')"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(/, 'hello')"
    # format-item-list
    for my_input in ["(/, 2(/'hello'))", "(/ 2(/'hello'))", "(/2(/'hello'))"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(/, 2(/, 'hello'))"
    # Hollerith string
    for my_input in ["(/, 3H2.2)", "(/ 3H2.2)", "(/3H2.2)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(/, 3H2.2)"

    # Comma is optional before or after a colon edit descriptor. There
    # are many combinations so just check with E : P.
    for my_input in ["(2E2.2, :, 2P)",
                     "(2E2.2 :, 2P)", "(2E2.2:, 2P)", "(2E2.2:,2P)",
                     "(2E2.2, : 2P)", "(2E2.2, :2P)", "(2E2.2,:2P)",
                     "(2E2.2 : 2P)", "(2E2.2: 2P)", "(2E2.2 :2P)",
                     "(2E2.2:2P)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(2E2.2, :, 2P)"


def test_c1002_clash(f2003_create):
    '''The constraints rules can clash with each other. Here we test what
    happens when they do. fparser assumes that we can break the P rule,
    as does PGI. In contrast gfortran allows P / but does not allow P
    : (which seems strange).

    '''
    # P followed by : with no comma. This breaks the P rule but is
    # satisfied by the : rule.
    my_input = "(2P :)"
    ast = Format_Specification(my_input)
    assert str(ast) == "(2P, :)"

    # P followed by / with no comma. This breaks the P rule but is
    # satisfied by the / rule.
    my_input = "(2P /)"
    ast = Format_Specification(my_input)
    assert str(ast) == "(2P, /)"


def test_c1002_triples(f2003_create):
    '''Test that we get expected behaviour when the C1002 rule applies to
    a triplet of items.

    '''
    for my_input in ["(2P, F2.2, /)", "(2P F2.2 /)", "(2PF2.2/)"]:
        ast = Format_Specification(my_input)
        assert str(ast) == "(2P, F2.2, /)"


def test_syntaxerror(f2003_create):
    ''' Test that we get an exception for incorrect bracket syntax. '''
    for my_input in [None, "", "  ", "(", ")", "x('hello')", "('hello')x"]:
        with pytest.raises(NoMatchError):
            _ = Format_Specification(my_input)


def test_syntaxerror_c1002(f2003_create):
    '''Test that we get an exception in situations where no comma is
    supplied and the C1002 constraints for optional commas do not
    apply.

    '''
    # Comma is mandatory before a slash edit descriptor when the
    # optional repeat specification is present
    # valid syntax
    for my_input in ["(2P, 2/)", "('hello', 2/)", "(2E2.2, 3/)"]:
        ast = Format_Specification(my_input)
    # invalid syntax
    for my_input in ["(2P 2/)", "(2P2/)", "('hello' 2/)", "('hello'2/)"]:
        with pytest.raises(NoMatchError):
            _ = Format_Specification(my_input)
    # Comma is mandatory after a P descriptor if not one of ['F', 'E',
    # 'EN', 'ES', 'D', 'G'] or not a '/' or a ':'
    # Test valid syntax.
    for descriptor in ['I', 'B', 'O', 'Z', 'L', 'A']:
        my_input = "(2P, {0}2)".format(descriptor)
        ast = Format_Specification(my_input)
        assert str(ast) == my_input
    # Test invalid syntax.
        for my_input in ["(2P {0}2)".format(descriptor),
                         "(2P{0}2)".format(descriptor)]:
            with pytest.raises(NoMatchError):
                _ = Format_Specification(my_input)
    # Check DT data descriptor
    my_input = "(2P, DT)"
    ast = Format_Specification(my_input)
    assert str(ast) == my_input
    my_input = "(2P DT)"
    with pytest.raises(NoMatchError):
        _ = Format_Specification(my_input)
    my_input = "(2PDT)"
    with pytest.raises(NoMatchError):
        _ = Format_Specification(my_input)
    # Comma is mandatory if C1002 is not relevant
    for my_input in ["('hello' 'hello')",
                     "(2P 2P)", "(2P2P)",
                     "(F2.2 F2.2)", "(F2.2F2.2)",
                     "(T2 T2)", "(T2T2)",
                     "(('hello') ('hello'))", "(('hello')('hello'))",
                     "(2H12 3H123)", "(2H123H123)"]:
        with pytest.raises(NoMatchError):
            _ = Format_Specification(my_input)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
##############################################################################
# Modified work Copyright (c) 2017 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
'''
Shared fixtures.
'''

import logging
import fparser.common.tests.logging_utils
import pytest


@pytest.fixture
def log():
    '''
    Prepare a fixture to capture logged events for inspection.
    '''
    logger = logging.getLogger('fparser')
    handler = fparser.common.tests.logging_utils.CaptureLoggingHandler()
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    yield handler
    # When the fixture function is a generator, i.e. it uses "yield" rather
    # than "return", it will be called a second time after the test has
    # completed. Thus anything appearing after the "yield" is teardown which
    # happens after the test.
    logger.removeHandler(handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#############################################################################
# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
'''
Tests the fparser.one.parsefortran module.
'''

import pytest
import fparser.one.parsefortran
import fparser.common.readfortran


def test_log_empty(log):
    '''
    Tests that a reader without next() method causes an event to be logged.
    '''
    class EmptyReader(object):
        '''
        A faux reader with no next() method.
        '''
        id = 'thingumy'

    unit_under_test = fparser.one.parsefortran.FortranParser(EmptyReader())
    unit_under_test.analyze()
    assert log.messages == {'debug':    [],
                            'info':     ['Nothing to analyze.'],
                            'warning':  [],
                            'error':    [],
                            'critical': []}


def test_log_cache(log):
    '''
    Tests that using a cached reader object logs an event.
    '''
    class Readerlike(object):
        '''
        Dummy reader class, the only purpose of which is to have an id and not
        cause the parser to fail.
        '''
        id = 'thisun'

        def next(self):
            '''
            Simple non-failure-causing method.
            '''
            yield 'NOT A THING'
            raise StopIteration

    # Expect everything to go okay, no log messages.
    log.reset()
    _ = fparser.one.parsefortran.FortranParser(Readerlike())
    assert log.messages == {'debug':    [],
                            'info':     [],
                            'warning':  [],
                            'error':    [],
                            'critical': []}

    # This time we should use a cached log.
    _ = fparser.one.parsefortran.FortranParser(Readerlike())
    assert log.messages == {'debug':    [],
                            'info':     ['using cached thisun'],
                            'warning':  [],
                            'error':    [],
                            'critical': []}


def test_log_failure(log, monkeypatch):
    '''
    Tests that an unexpected read failure causes an event to be logged.
    '''
    def faulty_next(self, ignore_comments=False):
        '''
        Raies any old exception.
        '''
        raise Exception('That''s all folks!')

    monkeypatch.setattr('fparser.common.readfortran.FortranStringReader.next',
                        faulty_next)
    reader = fparser.common.readfortran.FortranStringReader('')
    unit_under_test = fparser.one.parsefortran.FortranParser(reader)
    with pytest.raises(Exception):
        unit_under_test.parse()
    assert log.messages['debug'][0] \
        .startswith('An error occurred during parsing.')
    assert log.messages['info'] == []
    assert log.messages['warning'] == []
    assert log.messages['error'] == []
    assert log.messages['critical'][0].startswith('While processing')
    assert log.messages['critical'][1] == 'STOPPED PARSING'


def test_pyf():
    '''
    Tests inherited from implementation code.
    '''
    string = """
python module foo
  interface tere
    subroutine bar
    real r
    end subroutine bar
  end interface tere
end python module foo
"""
    expected = ['  PYTHONMODULE foo',
                '    INTERFACE tere',
                '      SUBROUTINE bar()',
                '        REAL r',
                '      END SUBROUTINE bar',
                '    END INTERFACE tere',
                '  END PYTHONMODULE foo']

    reader = fparser.common.readfortran.FortranStringReader(string)
    reader.set_format(fparser.common.sourceinfo.FortranFormat.from_mode('pyf'))
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()
    caught = parser.block.tofortran().splitlines()
    assert caught[0][:13] == '!BEGINSOURCE '
    assert caught[1:] == expected


def test_free90():
    '''
    Tests inherited from implementation code.
    '''
    string = """
module foo

   subroutine bar
    real r
    if ( pc_get_lun() .ne. 6) &
    write ( pc_get_lun(), '( &
    & /, a, /, " p=", i4, " stopping c_flag=", a, &
    & /, " print unit=", i8)') &
    trim(title), pcpsx_i_pel(), trim(c_flag), pc_get_lun()
    if (.true.) then
      call smth
    end if
    aaa : if (.false.) then
    else if (a) then
    else
    end if aaa
    hey = 1
    end subroutine bar
    abstract interface

    end interface

end module foo
"""
    expected = ['  MODULE foo',
                '    SUBROUTINE bar()',
                '      REAL r',
                '      IF (pc_get_lun() .ne. 6)' +
                ' WRITE (pc_get_lun(), \'(  /, A, /, " P = ", i4,' +
                ' " stopping c_flag=", a,  /, " print unit=", i8)\')' +
                ' trim(title), pcpsx_i_pel(), trim(c_flag), pc_get_lun()',
                '      IF (.true.) THEN',
                '        CALL smth',
                '      END IF',
                '      aaa: IF (.false.) THEN',
                '      ELSE IF (a) THEN',
                '      ELSE',
                '      END IF aaa',
                '      hey = 1',
                '    END SUBROUTINE bar',
                '    ABSTRACT INTERFACE',
                '    END INTERFACE',
                '  END MODULE foo']

    reader = fparser.common.readfortran.FortranStringReader(string)
    mode = fparser.common.sourceinfo.FortranFormat.from_mode('free')
    reader.set_format(mode)
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()
    caught = parser.block.tofortran().splitlines()
    assert caught[0][:13] == '!BEGINSOURCE '
    assert caught[1:] == expected


def test_module_procedure():
    '''
    Tests a type that contains a procedure, and makes sure
    it has a module_procedures attribute
    '''
    string = """
module foo
    type, public :: grid_type
   contains
        procedure :: get_tmask
   end type grid_type
end module foo
"""

    reader = fparser.common.readfortran.FortranStringReader(string)
    mode = fparser.common.sourceinfo.FortranFormat.from_mode('free')
    reader.set_format(mode)
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()

    # Fails if the class Type does not have a "module_procedures" attribute
    parser.analyze()


def test_f77():
    '''
    Tests inherited from implementation code.
    '''
    string = """\
      program foo
      a = 3
      end
      subroutine bar
      end
      pure function foo(a)
      end
      pure real*4 recursive function bar()
      end
"""
    expected = ['        PROGRAM foo',
                '          a = 3',
                '        END PROGRAM foo',
                '        SUBROUTINE bar()',
                '        END SUBROUTINE bar',
                '        pure FUNCTION foo(a)',
                '        END FUNCTION foo',
                '        pure recursive REAL*4 FUNCTION bar()',
                '        END FUNCTION bar']

    reader = fparser.common.readfortran.FortranStringReader(string)
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()
    assert isinstance(parser.block, fparser.one.block_statements.BeginSource)
    caught = parser.block.tofortran().splitlines()
    assert caught[0][:25] == '      !      BEGINSOURCE '
    assert caught[1:] == expected



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R312 : This file tests the support for an
extended intrinsic operator.

'''

import pytest
from fparser.two.Fortran2003 import Extended_Intrinsic_Op
from fparser.two.utils import NoMatchError


def test_extended_intrinsic_op(f2003_create):
    '''Check that correct extended intrinsic op input is parsed. No need
    to test them all as they will be tested in the pattern_tools tests
    but we do it anyway.

    '''
    for myinput in ["**", "*", "/", "+", "-", "//", ".eq.", ".NE.", ".lt.",
                    ".LE.", ".gt.", ".GE.", "==", "/=", "<", "<=", ">", ">=",
                    ".not.", ".AND.", ".or.", ".eQv.", ".NeqV."]:
        ast = Extended_Intrinsic_Op(myinput)
        assert myinput in str(ast)
        assert repr(ast) == "Extended_Intrinsic_Op('{0}')".format(myinput)


def test_parse_errors(f2003_create):
    '''The full set of errors will be checked in the pattern_tools
    tests. We just check a few here to make sure the correct
    NoMatchError exception is raised at this level.

    '''
    for myinput in ["", "  ", "***", "///", "eq.", ".eq", "eq", "= ="]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Extended_Intrinsic_Op(myinput)
        assert "Extended_Intrinsic_Op: '{0}'".format(myinput) in \
            str(excinfo.value)



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2017 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
# Modified M.Hambley, UK Met Office
##############################################################################
'''
Helps with testing methods which write to the standard logger.
'''
import logging


class CaptureLoggingHandler(logging.Handler):
    '''
    Records logged output for later examination.

    This is a standard handler for the built-in Python logging system.
    To make use of it simply register an instance with the logger using a
    command such as
    "logging.getLogger(__class__).addHandler( CaptureLoggingHandler() )"

    Any log message raised while this handler is in use will be recorded to
    a memory buffer for later use. This object has attributes 'debug', 'info',
    'warning', 'error' and 'critical', each of which is a list of logged
    messages. Only the message text is recorded, everything else is lost.
    '''
    def __init__(self, *args, **kwargs):
        '''
        Constructs an instance of CaptureLoggingHandler using the arguments
        accepted by logging.Handler.
        '''
        super(CaptureLoggingHandler, self).__init__(*args, **kwargs)
        self.reset()

    def emit(self, record):
        '''
        Handles a logged event. The event is passed from the logging system
        and recorded for future inspection.

        :param :py:class:`logging.LogRecord` record The event being logged.
        '''
        self.messages[record.levelname.lower()].append(record.getMessage())

    def reset(self):
        '''
        Empties the log of previous messages.
        '''
        self.messages = {'debug': [],
                         'info': [],
                         'warning': [],
                         'error': [],
                         'critical': []}



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R1227 : prefix.'''
import pytest

import fparser.two.Fortran2003 as f2003
from fparser.two.utils import NoMatchError


def test_prefix(f2003_create):
    '''Test that valid Prefix strings are matched successfully.

    '''
    # single space
    result = f2003.Prefix("impure elemental type(my_type) module")
    assert result.tostr() == "IMPURE ELEMENTAL TYPE(my_type) MODULE"
    assert (result.torepr() ==
            "Prefix(' ', (Prefix_Spec('IMPURE'), Prefix_Spec('ELEMENTAL'), "
            "Declaration_Type_Spec('TYPE', Type_Name('my_type')), "
            "Prefix_Spec('MODULE')))")

    # multiple spaces
    result = f2003.Prefix("  impure  elemental  type  (  my_type  )  module  ")
    assert result.tostr() == "IMPURE ELEMENTAL TYPE(my_type) MODULE"
    assert (result.torepr() ==
            "Prefix(' ', (Prefix_Spec('IMPURE'), Prefix_Spec('ELEMENTAL'), "
            "Declaration_Type_Spec('TYPE', Type_Name('my_type')), "
            "Prefix_Spec('MODULE')))")


def test_single_prefix_spec(f2003_create):
    '''Test that a single prefix-spec is returned as a Prefix containing a
    Prefix_Spec.

    '''
    result = f2003.Prefix("impure")
    assert result.tostr() == "IMPURE"
    assert (result.torepr() ==
            "Prefix(' ', (Prefix_Spec('IMPURE'),))")


def test_single_decl_spec(f2003_create):
    '''Test that a single prefix-spec is returned as a Prefix containing a
    declaration-type-spec.

    '''
    result = f2003.Prefix("type(my_type)")
    assert result.tostr() == "TYPE(my_type)"
    assert (result.torepr() ==
            "Prefix(' ', (Declaration_Type_Spec('TYPE', "
            "Type_Name('my_type')),))")


def test_decl_spec_lhs(f2003_create):
    '''Test that a Prefix containing a declaration-type-spec and
    prefix-spec keywords (e.g. IMPURE) to its left, produce the
    expected output.

    '''
    result = f2003.Prefix("impure recursive type(my_type)")
    assert result.tostr() == "IMPURE RECURSIVE TYPE(my_type)"
    assert (result.torepr() ==
            "Prefix(' ', (Prefix_Spec('IMPURE'), Prefix_Spec('RECURSIVE'), "
            "Declaration_Type_Spec('TYPE', Type_Name('my_type'))))")


def test_decl_spec_rhs(f2003_create):
    '''Test that a Prefix containing a declaration-type-spec and
    prefix-spec keywords (e.g. IMPURE) to its right, produce the
    expected output.

    '''
    result = f2003.Prefix("type(my_type) impure recursive")
    assert result.tostr() == "TYPE(my_type) IMPURE RECURSIVE"
    assert (result.torepr() ==
            "Prefix(' ', (Declaration_Type_Spec('TYPE', "
            "Type_Name('my_type')), Prefix_Spec('IMPURE'), "
            "Prefix_Spec('RECURSIVE')))")


def test_no_decl_spec(f2003_create):
    '''Test that a Prefix without a declaration-type-spec produces the
    expected output.

    '''
    result = f2003.Prefix("module impure")
    assert result.tostr() == "MODULE IMPURE"
    assert (result.torepr() ==
            "Prefix(' ', (Prefix_Spec('MODULE'), Prefix_Spec('IMPURE')))")


@pytest.mark.parametrize("string", ["invalid", "pure impure purile", "", " "])
def test_prefix_nomatch(f2003_create, string):
    '''Test that invalid Prefix strings raise a NoMatchError exception.

    '''
    with pytest.raises(NoMatchError):
        _ = f2003.Prefix(string)


def test_prefix_c1240(f2003_create):
    '''Test that an exception is raised if constraint c1240 is broken (at
    most one of each prefix-spec).

    '''
    with pytest.raises(NoMatchError):
        _ = f2003.Prefix("module module")

    with pytest.raises(NoMatchError):
        _ = f2003.Prefix("type(my_type) type(my_type2)")

    with pytest.raises(NoMatchError):
        _ = f2003.Prefix("module type(my_type2) module")


def test_prefix_c1241(f2003_create):
    '''Test that an exception is raised if constraint c1241 is broken
    (elemental and recursive are used together).

    '''
    with pytest.raises(NoMatchError):
        _ = f2003.Prefix("recursive elemental")



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2008 rule R1118 (C1113).

   parent-identifier is ancestor-module-name [ : parent-submodule-name ]

   C1113 The ancestor-module-name shall be the name of a nonintrinsic
   module; the parent-submodule name shall be the name of a
   descendant of that module.

   It is not possible to test C1113 with fparser as the names may be
   in different files.

'''
import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2008 import Parent_Identifier


def test_simple_1(f2008_create):
    '''Test the parsing of a minimal parent-identifier statement.'''
    result = Parent_Identifier("modulename")
    assert str(result) == "modulename"


def test_simple_2(f2008_create):
    '''Test the parsing of a minimal parent-identifier statement.'''
    result = Parent_Identifier("modulename : submodulename")
    assert str(result) == "modulename:submodulename"


def test_simple_error1(f2008_create):
    '''Test a parent_identifier statement with a syntax error raises an
    exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Parent_Identifier("modulename ; submodulename")
    assert ("Parent_Identifier: 'modulename ; submodulename'"
            in str(excinfo.value))


def test_simple_error2(f2008_create):
    '''Test a parent_identifier statement with a syntax error raises an
    exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Parent_Identifier("modulename :")
    assert "Parent_Identifier: 'modulename :'" in str(excinfo.value)


def test_simple_error3(f2008_create):
    '''Test a parent_identifier statement with a syntax error raises an
    exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Parent_Identifier(": submodulename")
    assert "Parent_Identifier: ': submodulename'" in str(excinfo.value)


def test_simple_error4(f2008_create):
    '''Test a parent_identifier statement with a syntax error raises an
    exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Parent_Identifier("modulename : submodulename : anothername")
    assert ("Parent_Identifier: 'modulename : submodulename : anothername'"
            in str(excinfo.value))


def test_simple_error5(f2008_create):
    '''Test a parent-identifier statement with additional content after
    the match raises an exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Parent_Identifier("modulename : submodulename :")
    assert ("Parent_Identifier: 'modulename : submodulename :'"
            in str(excinfo.value))



# Copyright (c) 2017-2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Test the various utility functions

"""
import io
import os

import pytest
import six

from fparser.common.utils import split_comma, ParseError, make_clean_tmpfile
from fparser.two.utils import InternalError


def test_split_comma():
    ''' Test the split_comma() function '''
    items = split_comma("hello, goodbye")
    print(items)
    assert items[0] == "hello"
    assert items[1] == "goodbye"
    # With trailing and leading white space
    items = split_comma("  hello, goodbye   ")
    print(items)
    assert items[0] == "hello"
    assert items[1] == "goodbye"
    items = split_comma("  ")
    assert not items


def test_split_comma_exceptions():
    ''' Test that we raise the expected exceptions if we don't supply
    the brackets in the right form '''
    with pytest.raises(ParseError) as excinfo:
        _ = split_comma("one, two", brackets="()")
    assert "brackets must be a tuple" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = split_comma("one, two", brackets=("()", ))
    assert "brackets tuple must contain just two items" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = split_comma("one, two", brackets=("(", "(", "("))
    assert "brackets tuple must contain just two items" in str(excinfo.value)


def test_split_bracketed_list():
    ''' Test the splitting of a list bracketed with parentheses '''
    items = split_comma("(well(1), this(is), it)", brackets=("(", ")"))
    print(items)
    assert items[0] == "well(1)"
    assert items[1] == "this(is)"
    # With superfluous white space
    items = split_comma("  (  well(1), this(is), it  )  ",
                        brackets=("(", ")"))
    print(items)
    assert items[0] == "well(1)"
    assert items[1] == "this(is)"
    assert items[2] == "it"
    # Square brackets
    items = split_comma("[well(1), this(is), it]", brackets=("[", "]"))
    print(items)
    assert items[0] == "well(1)"
    assert items[1] == "this(is)"
    assert items[2] == "it"
    # Mis-matched brackets
    items = split_comma("[well(1), this(is), it)", brackets=("[", "]"))
    assert not items


def test_extract_bracketed_list():
    ''' Test the extraction and parsing of a list within parentheses within
    a larger string '''
    from fparser.common.utils import extract_bracketed_list_items
    items = extract_bracketed_list_items("hello (this, is, a) test")
    assert items[0] == "this"
    assert items[1] == "is"
    assert items[2] == "a"


def test_extract_bracketed_list_err():
    ''' Test that we get the expected errors if the string passed into
    extract_bracketed_list_items() does not have the correct format '''
    from fparser.common.utils import extract_bracketed_list_items
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello (this, is, wrong(")
    assert "more than one opening/closing parenthesis found" in \
        str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello )this, is, wrong)")
    assert "more than one opening/closing parenthesis found" in \
        str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello (this, is, wrong) (too)")
    assert "more than one opening/closing parenthesis found" in \
        str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello )this, is, wrong( too")
    assert "failed to find expression within parentheses in" in \
        str(excinfo.value)

# utility for the make_clean_tmpfile tests


def create_tmp_file(string, tmpdir, filename="tmp_in.f90"):
    '''Utility function used by make_clean_tmpfile tests. Creates a
    temporary file and adds the content of a text string to it.

    :param str string: string containing content to be placed in a \
    file.
    :param tmpdir: pytest temporary directory structure.
    :type tmpdir: :py:class:'py._path.local.LocalPath'
    :param str filename: the name of the temporary file

    :returns: the filepath of the created file
    :rtype: str

    '''
    filepath = os.path.join(str(tmpdir), filename)
    # Create the input_file
    tmp_file = io.open(filepath, "w", encoding='UTF-8')
    if six.PY2:
        string = unicode(string)
    tmp_file.write(string)
    return tmp_file.name

# tests for the make_clean_tmpfile function


def test_mct_works(tmpdir):
    '''Test that if there are no errors in the input file then we get an
    exact copy of the input file in the temporary file.'''
    content = (
        "program valid\n"
        "end program valid\n")
    input_filepath = create_tmp_file(content, tmpdir)
    output_filepath = make_clean_tmpfile(input_filepath)
    with open(output_filepath, "r") as cfile:
        output = cfile.read()
    assert output == content


def test_mct_invalid_filename_arg():
    '''Test that the expected exception occurs if an invalid input file
    is provided.

    '''
    with pytest.raises(IOError) as excinfo:
        _ = make_clean_tmpfile("nonexistantfile")
    assert "No such file or directory" in str(excinfo.value)


def test_mct_invalid_skip_arg(tmpdir):
    '''Test that the expected exception occurs if an invalid
    skip_bad_input argument is provided.

    '''
    input_filepath = create_tmp_file("", tmpdir)
    with pytest.raises(InternalError) as excinfo:
        _ = make_clean_tmpfile(input_filepath, skip_bad_input="INVALID")
    assert ("skip_bad_input argument should be False or True but "
            "found 'INVALID'.") in str(excinfo.value)


def test_mct_invalid_encode_arg(tmpdir):
    '''Test that an appropriate exception is raised if the encode argument
    is invalid.

    '''
    input_filepath = create_tmp_file("", tmpdir)
    with pytest.raises(InternalError) as excinfo:
        _ = make_clean_tmpfile(input_filepath, encoding="invalid")
    assert "unknown encoding: invalid'." in str(excinfo.value)


def test_mct_parse_error(tmpdir):
    '''Test that an appropriate exception is raised if there is an invalid
    character in the input file and skip_bad_argument is set to
    False.

    '''
    invalid_content = u"\xca"
    input_filepath = create_tmp_file(invalid_content, tmpdir)
    with pytest.raises(ParseError) as excinfo:
        _ = make_clean_tmpfile(input_filepath, skip_bad_input=False,
                               encoding="ascii")
    assert ("Bad character in input file. Error returned was 'ascii' "
            "codec can't decode byte ") in str(excinfo.value)
    # Can't check the actual value as some versions of Python3 return
    # a different value to the one above.
    assert "in position 0: ordinal not in range(128)." in str(excinfo.value)


def test_mct_skip_error(tmpdir, caplog):
    '''Test that invalid characters are skipped in an input file by
    default and that logging messages are created.

    '''
    content = "HELLO"
    invalid_content = u"\xca".join(content)
    input_filepath = create_tmp_file(invalid_content, tmpdir)
    output_filepath = make_clean_tmpfile(input_filepath, encoding="ascii")
    with open(output_filepath, "r") as cfile:
        output = cfile.read()
    assert output == content
    for record in caplog.records:
        assert record.levelname != 'CRITICAL'
    assert ("Skipped bad character in input file. Error returned was 'ascii' "
            "codec can't decode byte ") in caplog.text
    # Can't check the actual value as some versions of Python3 return
    # a different value to the one above.
    assert "in position 1: ordinal not in range(128)." in caplog.text


def test_mct_utf8():
    '''Test that utf8 content can be read and written correctly in
    make_clean_tmpfile.

    '''
    filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            "utf.f90")
    _ = make_clean_tmpfile(filepath)



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 Cray-pointers: This file tests the support for a
Cray-pointee declaration.

'''

import pytest
from fparser.two.Fortran2003 import Cray_Pointee_Decl
from fparser.two.utils import NoMatchError


def test_cray_pointee_decl(f2003_create):
    '''Check that Cray-pointee declarations are parsed correctly.

    '''
    for myinput in ["a(n)", "a(0 : n)", "a(n, m)", "a(5, *)", "a(*)",
                    "a(0 : 1, 2 : *)"]:
        ast = Cray_Pointee_Decl(myinput)
        assert myinput in str(ast)


def test_errors(f2003_create):
    '''Check that syntax errors produce a NoMatchError exception.'''
    for myinput in ["", "  ", "a", "a2)", "a(2", "a()"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Cray_Pointee_Decl(myinput)
        assert "Cray_Pointee_Decl: '{0}'".format(myinput) in str(excinfo.value)


def test_unsupported(f2003_create):
    '''Check that unsupported assumed shape declarations produce a
    NoMatchError exception.

    '''
    for myinput in ["a(:)", "a(2,:)"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Cray_Pointee_Decl(myinput)
        assert "Cray_Pointee_Decl: '{0}'".format(myinput) in str(excinfo.value)



# Copyright (c) 2018-2019 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Module which provides pytest fixtures for use by files in this
directory

'''
import pytest
from fparser.two.parser import ParserFactory
from fparser.two.Fortran2003 import Defined_Unary_Op, Defined_Binary_Op


@pytest.fixture
def f2003_create():
    '''Create a fortran 2003 parser class hierarchy'''
    _ = ParserFactory().create(std="f2003")


@pytest.fixture(scope="module", params=[Defined_Unary_Op, Defined_Binary_Op])
def op_type(request):
    '''Fixture for testing the two types of defined op (unary and
    binary).

    '''
    return request.param



# Modified work Copyright (c) 2017-2019 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

'''
Provides functions to determine whether a piece of Fortran source is free or
fixed format. It also tries to differentiate between strict and "pyf" although
I'm not sure what that is.
'''

import io
import os
import re
import six


##############################################################################

class FortranFormat(object):
    '''
    Describes the nature of a piece of Fortran source.

    Source can be fixed or free format. It can also be "strict" or
    "not strict" although it's not entirely clear what that means. It may
    refer to the strictness of adherance to fixed format although what that
    means in the context of free format I don't know.
    '''
    def __init__(self, is_free, is_strict):
        '''
        Constructs a FortranFormat object from the describing parameters.

        Arguments:
            is_free   - (Boolean) True for free format, False for fixed.
            is_strict - (Boolean) Some amount of strictness.
        '''
        if is_free is None:
            raise Exception('FortranFormat does not accept a None is_free')
        if is_strict is None:
            raise Exception('FortranFormat does not accept a None is_strict')

        self._is_free = is_free
        self._is_strict = is_strict

    @classmethod
    def from_mode(cls, mode):
        '''
        Constructs a FortranFormat object from a mode string.

        Arguments:
            mode - (String) One of 'free', 'fix', 'f77' or 'pyf'
        '''
        if mode == 'free':
            is_free, is_strict = True, False
        elif mode == 'fix':
            is_free, is_strict = False, False
        elif mode == 'f77':
            is_free, is_strict = False, True
        elif mode == 'pyf':
            is_free, is_strict = True, True
        else:
            raise NotImplementedError(repr(mode))
        return cls(is_free, is_strict)

    def __eq__(self, other):
        if isinstance(other, FortranFormat):
            return self.is_free == other.is_free \
                   and self.is_strict == other.is_strict
        raise NotImplementedError

    def __str__(self):
        if self.is_strict:
            string = 'Strict'
        else:
            string = 'Non-strict'

        if self.is_free:
            string += ' free'
        else:
            string += ' fixed'

        return string + ' format'

    @property
    def is_free(self):
        '''
        Returns true for free format.
        '''
        return self._is_free

    @property
    def is_fixed(self):
        '''
        Returns true for fixed format.
        '''
        return not self._is_free

    @property
    def is_strict(self):
        '''
        Returns true for strict format.
        '''
        return self._is_strict

    @property
    def is_f77(self):
        '''
        Returns true for strict fixed format.
        '''
        return not self._is_free and self._is_strict

    @property
    def is_fix(self):
        '''
        Returns true for slack fixed format.
        '''
        return not self._is_free and not self._is_strict

    @property
    def is_pyf(self):
        '''
        Returns true for strict free format.
        '''
        return self._is_free and self._is_strict

    @property
    def mode(self):
        '''
        Returns a string representing this format.
        '''
        if self._is_free and self._is_strict:
            mode = 'pyf'
        elif self._is_free:
            mode = 'free'
        elif self.is_fix:
            mode = 'fix'
        elif self.is_f77:
            mode = 'f77'
        # While mode is determined by is_free and is_strict all permutations
        # are covered. There is no need for a final "else" clause as the
        # object cannot get wedged in an invalid mode.
        return mode


##############################################################################

_HAS_F_EXTENSION = re.compile(r'.*[.](for|ftn|f77|f)\Z', re.I).match

_HAS_F_HEADER = re.compile(r'-[*]-\s*(fortran|f77)\s*-[*]-', re.I).search
_HAS_F90_HEADER = re.compile(r'-[*]-\s*f90\s*-[*]-', re.I).search
_HAS_F03_HEADER = re.compile(r'-[*]-\s*f03\s*-[*]-', re.I).search
_HAS_F08_HEADER = re.compile(r'-[*]-\s*f08\s*-[*]-', re.I).search
_HAS_FREE_HEADER = re.compile(r'-[*]-\s*(f90|f95|f03|f08)\s*-[*]-',
                              re.I).search
_HAS_FIX_HEADER = re.compile(r'-[*]-\s*fix\s*-[*]-', re.I).search
_HAS_PYF_HEADER = re.compile(r'-[*]-\s*pyf\s*-[*]-', re.I).search

_FREE_FORMAT_START = re.compile(r'[^c*!]\s*[^\s\d\t]', re.I).match


def get_source_info_str(source):
    '''
    Determines the format of Fortran source held in a string.

    Returns a FortranFormat object.
    '''
    lines = source.splitlines()
    if not lines:
        return FortranFormat(False, False)

    firstline = lines[0].lstrip()
    if _HAS_F_HEADER(firstline):
        return FortranFormat(False, True)
    if _HAS_FIX_HEADER(firstline):
        return FortranFormat(False, False)
    if _HAS_FREE_HEADER(firstline):
        return FortranFormat(True, False)
    if _HAS_PYF_HEADER(firstline):
        return FortranFormat(True, True)

    line_tally = 10000  # Check up to this number of non-comment lines
    is_free = False
    while line_tally > 0 and lines:
        line = lines.pop(0).rstrip()
        if line and line[0] != '!':
            line_tally -= 1
            if line[0] != '\t' and _FREE_FORMAT_START(line[:5]) \
               or line[-1:] == '&':
                is_free = True
                break

    return FortranFormat(is_free, False)


##############################################################################

def get_source_info(file_candidate):
    '''
    Determines the format of Fortran source held in a file.

    :param file_candidate: a filename or a file object
    :type file_candidate: str or (file (py2) or _io.TextIOWrapper (py3))

    :returns: the Fortran format encoded as a string.
    :rtype: str

    '''
    if hasattr(file_candidate, 'name') and hasattr(file_candidate, 'read'):
        filename = file_candidate.name

        # The behaviour of file.name when associated with a file without a
        # file name has changed between Python 2 and 3.
        #
        # Under Python 3 file.name holds an integer file handle.
        if isinstance(filename, int):
            filename = None

        # Under Python 2 file.name holds a string of the form "<..>".
        elif filename.startswith('<') and filename.endswith('>'):
            filename = None
    elif isinstance(file_candidate, six.string_types):
        # The preferred method for identifying strings changed between Python2
        # and Python3.
        filename = file_candidate
    else:
        message = 'Argument must be a filename or file-like object.'
        raise ValueError(message)

    if filename:
        _, ext = os.path.splitext(filename)
        if ext == '.pyf':
            return FortranFormat(True, True)

    if hasattr(file_candidate, 'read'):
        # If the candidate object has a "read" method we assume it's a file
        # object.
        #
        # If it is a file object then it may be in the process of being read.
        # As such we need to take a note of the current state of the file
        # pointer so we can restore it when we've finished what we're doing.
        #
        pointer = file_candidate.tell()
        file_candidate.seek(0)
        source_info = get_source_info_str(file_candidate.read())
        file_candidate.seek(pointer)
        return source_info
    else:
        # It isn't a file and it passed the type check above so it must be
        # a string.
        #
        # If it's a string we assume it is a filename. In which case we need
        # to open the named file so we can read it.
        #
        # It is closed on completion so as to return it to the state it was
        # found in.
        #
        from fparser.common.utils import make_clean_tmpfile
        tmpfile = make_clean_tmpfile(file_candidate)
        with io.open(tmpfile, 'r', encoding='utf8') as file_object:
            string = get_source_info_str(file_object.read())
        os.remove(tmpfile)
        return string

##############################################################################



# Copyright (c) 2018-2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R1013 : This file tests support for the
Position_Edit_Desc class.

'''

import pytest
from fparser.two.utils import NoMatchError, InternalError
from fparser.two.Fortran2003 import Position_Edit_Desc


def test_invalid_descriptor():
    '''Test invalid options raise an exception.'''

    for descriptor in [None, "", "  ", "1T", "XT", "T", "TL", "TR", "XT1",
                       "XX", "X X", "1XX"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Position_Edit_Desc(descriptor)
        assert "Position_Edit_Desc: '{0}'".format(descriptor) \
            in str(excinfo.value)


def test_valid_t_descriptor(f2003_create):
    '''Test valid nT, nTL and nTR inputs provide the expected output'''

    for name in ["T", "TL", "TR"]:
        for descriptor in ["{0}1".format(name),
                           "  {0}  1  ".format(name),
                           "{0}999".format(name),
                           "{0}999".format(name.lower())]:
            result = Position_Edit_Desc(descriptor)
            assert str(result) == "".join(descriptor.split()).upper()


def test_valid_x_descriptor(f2003_create):
    '''Test valid nX inputs provide the expected output'''

    for descriptor in ["1X", "  1  X  ", "999X", "999x"]:
        result = Position_Edit_Desc(descriptor)
        assert str(result) == "".join(descriptor.split()).upper()


def test_invalid_x_descriptor1(f2003_create, monkeypatch):
    '''Test that the X extension to the standard raises an exception if it
    is not named as a valid extension.

    '''

    import fparser.two.utils as utils
    monkeypatch.setattr(utils, "EXTENSIONS", [])
    for descriptor in ["X", "  X  ", "x"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Position_Edit_Desc(descriptor)
        assert "Position_Edit_Desc: '{0}'".format(descriptor) \
            in str(excinfo.value)


def test_valid_x_descriptor2(f2003_create, monkeypatch):
    '''Test that the X extension to the standard produces the expected
    output if it is named as a valid extension.

    '''

    import fparser.two.utils as utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["x-format"])
    for descriptor in ["X", "  X  ", "x"]:
        result = Position_Edit_Desc(descriptor)
        assert str(result) == "X"


def test_tostr_invalid_1(f2003_create, monkeypatch):
    '''Test that invalid input (the items list is not the expected size)
    raises an exception in the tostr method

    '''

    obj = Position_Edit_Desc("1X")
    monkeypatch.setattr(obj, "items", [])
    with pytest.raises(InternalError) as excinfo:
        _ = str(obj)
    assert "tostr() has '0' items, but expecting 2" in str(excinfo.value)


def test_tostr_invalid_2(f2003_create, monkeypatch):
    '''Test that invalid input (items[1] is empty) raises an exception in
    the tostr method'''

    obj = Position_Edit_Desc("1X")
    monkeypatch.setattr(obj, "items", ["1", ""])
    with pytest.raises(InternalError) as excinfo:
        _ = str(obj)
    assert ("items[1] in Class Position_Edit_Desc method tostr() is "
            "empty or None") in str(excinfo.value)



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2008 rule R1119.

    end-submodule-stmt is END [ SUBMODULE [ submodule-name ] ]

'''
import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2008 import End_Submodule_Stmt


def test_simple_1(f2008_create):
    '''Test the parsing of a minimal end-submodule statement.'''
    result = End_Submodule_Stmt("end")
    assert str(result) == "END"


def test_simple_2(f2008_create):
    '''Test the parsing of an end-submodule statement which includes the
    submodule keyword.

    '''
    result = End_Submodule_Stmt("end submodule")
    assert str(result) == "END SUBMODULE"


def test_simple_3(f2008_create):
    '''Test the parsing of an end-submodule statement which includes the
    submodule name

    '''
    result = End_Submodule_Stmt("end submodule name")
    assert str(result) == "END SUBMODULE name"


def test_simple_error1(f2008_create):
    '''Test an end-submodule statement with a syntax error raises an
    exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = End_Submodule_Stmt("edn")
    assert "End_Submodule_Stmt: 'edn'" in str(excinfo.value)


def test_simple_error2(f2008_create):
    '''Test an end-submodule statement with a syntax error raises an
    exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = End_Submodule_Stmt("end submod")
    assert "End_Submodule_Stmt: 'end submod'" in str(excinfo.value)


def test_simple_error3(f2008_create):
    '''Test an end-submodule statement with additional content after the
    match raises an exception.

    '''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = End_Submodule_Stmt("end submodule name :")
    assert "End_Submodule_Stmt: 'end submodule name :'" in str(excinfo.value)



# Copyright (c) 2018-2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R201 : This file tests the support for zero
or more program-units.

'''

import pytest
from fparser.two.utils import FortranSyntaxError
from fparser.api import get_reader
from fparser.two.Fortran2003 import Program

# Test no content or just white space. This is not officially a
# Fortran rule but fortran compilers tend to accept empty content so
# we follow their lead.


def test_empty_input(f2003_create):
    '''Test that empty input or input only containing white space can be
    parsed succesfully

    '''
    for code in ["", "   ", "  \n  \n\n"]:
        reader = get_reader(code)
        ast = Program(reader)
        assert str(ast) == ""

# Test single program units


@pytest.mark.usefixtures("f2003_create")
def test_single():
    '''Test that a single program_unit can be parsed successfully.'''
    reader = get_reader('''\
      subroutine test()
      end subroutine
      ''')
    ast = Program(reader)
    assert "SUBROUTINE test\n" \
        "END SUBROUTINE" in str(ast)
    # Check that the Name of the subroutine has the correct parent
    assert ast.content[0].content[0].items[1].parent is \
        ast.content[0].content[0]


@pytest.mark.usefixtures("f2003_create")
def test_single_with_end_name():
    '''Test that a single program_unit can be parsed successfully when it
    has a name specified on the end clause.'''
    reader = get_reader('''\
      subroutine test()
      end subroutine test
      ''')
    ast = Program(reader)
    assert "SUBROUTINE test\n" \
        "END SUBROUTINE test" in str(ast)
    # Check parent information has been set-up correctly
    end_sub = ast.content[0].content[-1]
    assert end_sub.items[1].parent is end_sub


@pytest.mark.xfail(reason="5 spaces causes the error exception to occur at "
                   "the end of the file")
def test_single2(f2003_create):
    '''Test that a single program_unit with 5 or more spaces at the start
    of the line reports an error on the correct (first) line

    '''
    reader = get_reader('''\
     subroutin test()
     end subroutine

      ''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("at line 1\n>>>      subroutin test()\n"
            in str(excinfo.value))


@pytest.mark.xfail(reason="5 spaces causes the error exception to occur at "
                   "the end of the file")
def test_single3(f2003_create):
    '''Test that a single program_unit with 5 or more spaces at the start
    of the line reports an error on the correct (second) line

    '''
    reader = get_reader('''\
     subroutine test()
     end subroutin

      ''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("at line 2\n>>>      end subroutin\n"
            in str(excinfo.value))


def test_single_error1(f2003_create):
    '''Test that a single program_unit with an error in the initial
    statement raises an appropriate exception

    '''
    reader = get_reader('''\
      subroutin test()
      end subroutine
      ''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("at line 1\n>>>      subroutin test()\n"
            in str(excinfo.value))


def test_single_error2(f2003_create):
    '''Test that a single program_unit with an error in the final
    statement raises an appropriate exception

    '''
    reader = get_reader(
        "subroutine test()\n\n"
        "end subroutin\n\n\n")
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy = Program(reader)
    assert ("at line 3\n>>>end subroutin\n"
            in str(excinfo.value))

# Test multiple program units


def test_multiple(f2003_create):
    '''Test that multiple program_units can be parsed successfully.'''
    reader = get_reader('''\
      subroutine test()
      end subroutine
      subroutine test2()
      end subroutine test2
      ''')
    ast = Program(reader)
    assert "SUBROUTINE test\n" \
        "END SUBROUTINE" in str(ast)


@pytest.mark.xfail(reason="Having the same name in different program_units "
                   "does not raise an exception")
def test_multiple_error1(f2003_create):
    '''Test that multiple program_units with the same name raise an
    exception

    '''
    reader = get_reader('''\
      subroutine test()
      end subroutine
      subroutine test()
      end subroutine
      ''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("XXX"
            in str(excinfo.value))


def test_multiple_error2(f2003_create):
    '''Test that a single program_unit with an error raises an appropriate
    exception

    '''
    reader = get_reader('''\
      subroutine 1test()
      end subroutine
      subroutine test()
      end subroutine
      ''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("at line 1\n>>>      subroutine 1test()\n"
            in str(excinfo.value))


def test_multiple_error3(f2003_create):
    '''Test that multiple program_units with an error raises an
    appropriate exception

    '''
    reader = get_reader('''\
      subroutine test()
      end subroutine
      subroutine test()
      end subroutin''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("at line 4\n>>>      end subroutin\n"
            in str(excinfo.value))

# Test a program unit with a missing program statement


def test_missing_prog(f2003_create):
    '''Test that a main program program_unit without a program declaration
    can be parsed successfully. This should not really be a test here,
    but this case is currently treated separately by the match method
    in Program.

    '''
    reader = get_reader('''\
      end
      ''')
    ast = Program(reader)
    assert "END" in str(ast)


@pytest.mark.xfail(reason="Only the main program is output")
def test_missing_prog_multi(f2003_create):
    '''Test that a main program program_unit without a program declaration
    can be parsed successfully when it is not the first program_unit.

    '''
    reader = get_reader('''\
      subroutine first
      end
      end
      ''')
    ast = Program(reader)
    assert "SUBROUTINE first\n" \
        "END SUBROUTINE" in str(ast)
    assert "END PROGRAM" in str(ast)

# A program should contain only only one main program


@pytest.mark.xfail(reason="Only one main program is allowed in a program")
def test_one_main1(f2003_create):
    '''Test that multiple main programs raise an exception.'''
    reader = get_reader('''\
      program first
      end
      program second
      end
      ''')
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert ("XXX"
            in str(excinfo.value))


# Check comments are supported at this level


def test_comment0(f2003_create):
    '''Test that a single program_unit without comments can be parsed
    successfully with comment processing switched on.

    '''
    reader = get_reader((
        "subroutine test()\n"
        "end subroutine\n"), ignore_comments=False)
    ast = Program(reader)
    assert ("SUBROUTINE test\n"
            "END SUBROUTINE") in str(ast)


def test_comment1(f2003_create):
    '''Test that a single program_unit can be parsed successfully with
    comments being ignored.'''
    reader = get_reader('''\
      ! comment1
      subroutine test()
      end subroutine
      ! comment2
      ''')
    ast = Program(reader)
    assert "SUBROUTINE test\n" \
        "END SUBROUTINE" in str(ast)
    assert "! comment" not in str(ast)


def test_comment2(f2003_create):
    '''Test that a single program_unit can be parsed successfully with
    comments being included.'''
    reader = get_reader((
        "! comment1\n"
        "subroutine test()\n"
        "end subroutine\n"
        "! comment2\n"), ignore_comments=False)
    ast = Program(reader)
    assert ("! comment1\n"
            "SUBROUTINE test\n"
            "END SUBROUTINE\n"
            "! comment2") in str(ast)


def test_comment3(f2003_create):
    '''Test that multiple program_units can be parsed successfully with
    comments being ignored.'''
    reader = get_reader('''\
      ! comment1
      subroutine test()
      end subroutine
      ! comment2
      module example
      end module
      ! comment3
      ''', ignore_comments=True)
    ast = Program(reader)
    assert ("SUBROUTINE test\n"
            "END SUBROUTINE\n"
            "MODULE example\n"
            "END MODULE") in str(ast)
    assert "! comment" not in str(ast)


def test_comment4(f2003_create):
    '''Test that multiple program_units can be parsed successfully with
    comments being included.'''
    reader = get_reader((
        "! comment1\n"
        "subroutine test()\n"
        "end subroutine\n"
        "! comment2\n"
        "module example\n"
        "end module\n"
        "! comment3\n"
    ), ignore_comments=False)
    ast = Program(reader)
    assert ("! comment1\n"
            "SUBROUTINE test\n"
            "END SUBROUTINE\n"
            "! comment2\n"
            "MODULE example\n"
            "END MODULE\n"
            "! comment3") in str(ast)

# Check includes are supported at this level


def test_include0(f2003_create):
    '''Test that a single program_unit with includes can be parsed
    succesfully.

    '''
    reader = get_reader((
        "include '1'\n"
        "subroutine test()\n"
        "end subroutine\n"
        "include '2'\n"))
    ast = Program(reader)
    assert ("INCLUDE '1'\n"
            "SUBROUTINE test\n"
            "END SUBROUTINE\n"
            "INCLUDE '2'") in str(ast)


def test_include1(f2003_create):
    '''Test that multiple program_units with includes can be parsed
    successfully.

    '''
    reader = get_reader(
        "include '1'\n"
        "subroutine test()\n"
        "end subroutine\n"
        "include '2'\n"
        "module example\n"
        "end module\n"
        "include '3'\n", ignore_comments=True)
    ast = Program(reader)
    assert ("INCLUDE '1'\n"
            "SUBROUTINE test\n"
            "END SUBROUTINE\n"
            "INCLUDE '2'\n"
            "MODULE example\n"
            "END MODULE\n"
            "INCLUDE '3'") in str(ast)
    assert "! comment" not in str(ast)

# Check a mix of includes and comments are supported at this level


def test_mix(f2003_create):
    '''Test that multiple program_units can be parsed successfully with a
    mix of includes and comments.

    '''
    reader = get_reader((
        "include '1'\n"
        "! comment1\n"
        "include '2'\n"
        "subroutine test()\n"
        "end subroutine\n"
        "include '3'\n"
        "include '4'\n"
        "! comment2\n"
        "! comment3\n"
        "module example\n"
        "end module\n"
        "! comment4\n"
        "include '5'\n"
        "! comment5\n"
    ), ignore_comments=False)
    ast = Program(reader)
    assert ("INCLUDE '1'\n"
            "! comment1\n"
            "INCLUDE '2'\n"
            "SUBROUTINE test\n"
            "END SUBROUTINE\n"
            "INCLUDE '3'\n"
            "INCLUDE '4'\n"
            "! comment2\n"
            "! comment3\n"
            "MODULE example\n"
            "END MODULE\n"
            "! comment4\n"
            "INCLUDE '5'\n"
            "! comment5") in str(ast)



# This Python file uses the following encoding: utf-8
# Copyright (c) 2019-2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' pytest module for the Fortran2003 Case Construct - R808.
    Does not test all aspects of R808, in particular the conditions C803-7
    are not checked - #232. '''

import pytest
from fparser.api import get_reader
from fparser.common.readfortran import FortranStringReader
from fparser.two.Fortran2003 import Case_Construct


@pytest.mark.usefixtures("f2003_create")
def test_case_construct():
    ''' Basic test that we parse a Case Construct successfully. '''
    tcls = Case_Construct
    obj = tcls(get_reader('''\
select case (n)
case (:-1)
  signum = -1
case (0)
  signum = 0
case (1:)
  signum = 1
case default
  signum = -2
end select
'''))
    assert isinstance(obj, tcls), repr(obj)
    assert (str(obj) ==
            'SELECT CASE (n)\nCASE (: - 1)\n  signum = - 1\nCASE (0)\n'
            '  signum = 0\nCASE (1 :)\n  signum = 1\nCASE DEFAULT\n'
            '  signum = - 2\nEND SELECT')


@pytest.mark.usefixtures("f2003_create")
def test_tofortran_non_ascii():
    ''' Check that the tofortran() method works when the character string
    contains non-ascii characters. '''
    code = (u"SELECT CASE(iflag)\n"
            u"CASE(  30  ) ! This is a comment\n"
            u"  IF(lwp) WRITE(*,*) ' for e1=1\xb0'\n"
            u"END SELECT\n")
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Case_Construct(reader)
    out_str = str(obj)
    assert "for e1=1" in out_str



#!/usr/bin/env python
# Modified work Copyright (c) 2017-2019 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

''' Example script to parse a Fortran program using fparser '''
from __future__ import print_function
import logging
import sys
from fparser.scripts.script_options import set_fparser_options

logging.basicConfig()

try:
    from iocbio.optparse_gui import OptionParser
except ImportError:
    from optparse import OptionParser


def runner(_, options, args):
    '''
    Function to read, parse and output Fortran source code.

    :param options: object constructed by OptionParser with cmd-line flags.
    :param args: list of Fortran files to parse.
    :type args: list of str

    '''
    import six
    from fparser.two.parser import ParserFactory
    from fparser.two.Fortran2003 import FortranSyntaxError, InternalError
    from fparser.common.readfortran import FortranFileReader
    if not args:
        print("Error: No fortran files specified", file=sys.stderr)
        raise SystemExit(1)
    for filename in args:
        print("File: '{0}'".format(filename), file=sys.stderr)
        try:
            reader = FortranFileReader(filename,
                                       ignore_comments=False)
        except IOError as error:
            print(error, file=sys.stderr)
            continue
        try:
            fparser = ParserFactory().create(std=options.std)
            program = fparser(reader)
            if options.task == "show":
                print(six.text_type(program))
            if options.task == "repr":
                print(repr(program))
        except FortranSyntaxError as msg:
            print("Syntax error: {0}".format(six.text_type(msg)),
                  file=sys.stderr)
        except InternalError as msg:
            print("Internal error in fparser: {0}".format(six.text_type(msg)),
                  file=sys.stderr)


def main():
    ''' Check arguments before parsing code '''
    parser = OptionParser()
    set_fparser_options(parser)
    options, args = parser.parse_args()
    runner(parser, options, args)


if __name__ == "__main__":
    main()  # pragma: no cover
